<?php
// **************************************************************** //
// These are Square Connect Platform Functions Created by Ryan May //
// **************************************************************** //


function get_oauth_url($payable_user = false, $oauth_environment = 'testing', $source = false){

    global $SQUARE_CONFIG, $environment;

    $square_scopes = array(
        'PAYMENTS_WRITE',
        'PAYMENTS_READ',
        'PAYMENTS_WRITE_ADDITIONAL_RECIPIENTS',
        'MERCHANT_PROFILE_READ',
        'DEVICE_CREDENTIAL_MANAGEMENT',
        'CUSTOMERS_READ',
        'SUBSCRIPTIONS_WRITE',
        'ITEMS_READ',
        'ORDERS_WRITE',
        'ORDERS_READ',
        'INVOICES_WRITE',
        'CUSTOMERS_READ',
        'CUSTOMERS_WRITE'
    );


    //This function whill create a square OAUTH URL.
    $url_base = 'https://'.$SQUARE_CONFIG[$oauth_environment]['api_domain'];
    $url_base .='/oauth2/authorize?';
    $url_base .='session=False';
    $url_base .='&scope='.implode('+', $square_scopes);
    $oauth_url = $url_base . '&client_id='.$SQUARE_CONFIG[$oauth_environment]['client_id'];

    return $oauth_url;

}


function get_merchant_details($square_merchant_id, $access_token, $environment = 'testing'){

    global $SQUARE_CONFIG;

    $curl = curl_init();

    curl_setopt_array($curl, array(
      CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$environment]['api_domain'].'/v2/merchants/'.$square_merchant_id,
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'GET',
      CURLOPT_HTTPHEADER => array(
        'Authorization: Bearer '.$access_token
      ),
    ));

    $response = curl_exec($curl);

    $info = curl_getinfo($curl);

    curl_close($curl);


    $respons_array = json_decode($response, true);

    return $respons_array;

}


function get_square_api_credentials(){

    global $DB_LINK, $SQUARE_CONFIG;

    $credentials = array();

    //Lets lookup the connection ID
    if(isset($_COOKIE['cashEAR']) && !empty($_COOKIE['cashEAR'])){

        $credentials_array = json_decode(base64_decode($_COOKIE['cashEAR']), true);
        $credentials = $credentials_array;


        //before we return this - Lets check if the access token is expired.
        if (strtotime($credentials_array['expires_at']) < strtotime('+1 days')){ //looks like it expires in a day so lets refresh it.

            //So lets fix that and refresh the access token
            $curl = curl_init();

            curl_setopt_array($curl, array(
              CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$environment]['api_domain'].'/oauth2/token',
              CURLOPT_RETURNTRANSFER => true,
              CURLOPT_ENCODING => '',
              CURLOPT_MAXREDIRS => 10,
              CURLOPT_TIMEOUT => 0,
              CURLOPT_FOLLOWLOCATION => true,
              CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
              CURLOPT_CUSTOMREQUEST => 'POST',
              CURLOPT_POSTFIELDS =>'{
                "client_id": "'.$SQUARE_CONFIG[$environment][$app_region]['client_id'].'",
                "client_secret": "'.$SQUARE_CONFIG[$environment][$app_region]['client_secret'].'",
                "refresh_token": "'.$credentials['refresh_token'].'",
                "grant_type": "refresh_token"
              }',
              CURLOPT_HTTPHEADER => array(
                'Square-Version: 2021-09-21',
                'Content-Type: application/json'
              ),
            ));

            $response = curl_exec($curl);
            $info = curl_getinfo($curl);

            curl_close($curl);

            $refreshed_credentials_array = json_decode($response, true);

            if(isset($refreshed_credentials_array['access_token'])){

                //We got updated credentials, so lets merche it with the other merchant info 
                $credentials = array_merge($credentials, $refreshed_credentials_array);
                //lets save this new setup in the DB with the connection
                $new_encoded_credentials = base64_encode(json_encode($credentials));

                if(mysqli_query($DB_LINK, "UPDATE `payment_provider_connections` SET `configuration_vars` = '$new_encoded_credentials', `updated` = NOW() WHERE `payment_provider_connections`.`id` = $connection_id;")){

                }else{
                    //Hmm we failed to save the new credentials.
                    //save_log('provider_connection:'.$connection_id, 'update_config_vars', $new_encoded_credentials, 'Failed to save new vars in DB', 500, 0);
                }


            }else{ //hmm thats not good.
                $credentials = false;
            }


        }else{ //Not expired all is well

        }


    }else{ //looks like there is no connection yet.. 

        $credentials = false;

        //TODO use the sahred credentials.

    }

    return $credentials;

}



function create_order_from_payable_id($payable_order_id){  //This will create a Square Order Object.


    global $SQUARE_CONFIG, $DB_LINK, $environment;

    $response = array();
    $response['message'] = 'Initalizing';
    $response['success'] = false;

    // Step 1 is to load the order from the database
    $order_details = get_order($payable_order_id, true);
    $internal_order_id = $order_details['order']['id'];

    //Let's set a string we can use to specify the credentials.
    if($order_details['order']['test_mode']){
        $square_env_label = 'testing';
    }else{
        $square_env_label = 'production';
    }

    $response['order_environment'] = $square_env_label;

    $square_credentials = get_square_api_credentials($order_details['source']['payment_provider_connection_id'], $square_env_label);

    $payable_application_fee = calculate_platform_fee($order_details['order']['total'], $order_details['order']['order_currency'], date("Y"));

    //TODO - Its smallest currency unit, some currencies may not need to be multiplied.
    $square_total_amount = round(floatval($order_details['order']['total'])*100); //needs to be in cents
    $square_application_fee = round(floatval($payable_application_fee['fee']['amount'])*100); //needs to be in cents

    $api_request_id = generate_key(20);

    $square_order_data = [
      'idempotency_key' => $api_request_id,
      'order' => [
          'location_id' => $square_credentials['merchant']['main_location_id'],
          'reference_id' => $payable_order_id,
          'source' => ['name' => 'Google Form: '.$order_details['source']['g_form_title']],
          'line_items' => [],
          'metadata' => [
            'payable_order_id' => $payable_order_id,
            'payable_source_plugin' => $order_details['order']['source_plugin'],
            'payable_source_id' => $order_details['source']['payable_id'],
          ]
      ]
    ];

    //We have to loop thorugh the line items and add them to the Request.
    $line_items = json_decode($order_details['order']['line_items'], true);

    foreach($line_items as $an_item){

        if(empty($an_item['item_name'])){
            $an_item['item_name'] = "Item";
        }

        $square_formatted_item = [

            'name' => $an_item['item_name'],
            'quantity' => '1',
            'base_price_money' => [
                'currency' => $order_details['order']['order_currency'],
                'amount' => round(floatval($an_item['amount'])*100) //needs to be in cents
            ]
        ];

        array_push($square_order_data['order']['line_items'], $square_formatted_item);
    }

    if(!empty($order_details['order']['tax_total'])){ //we have some taxes to add

        $square_order_data['order']['taxes'][0] = [
            'name' => 'TAX',
            'type' => 'ADDITIVE',
            'percentage' => strval($order_details['source']['add_taxes']*100),
            'applied_money' => [
                'currency' => $order_details['order']['order_currency'],
                'amount' => round(floatval($order_details['order']['tax_total'])*100) //needs to be in cents
            ]
        ];

    } //end if we have to add taxes


    if(!empty($order_details['order']['handling_total'])){ //If we have a handling fee we need to add it.

        $square_order_data['order']['service_charges'][0] = [
            'name' => 'Handling and Convenience Fee',
            'percentage' => strval($order_details['source']['add_convenience_fee']*100),
            'taxable' => false,
            'calculation_phase' => 'TOTAL_PHASE'
        ];

    } //end if we have to add taxes

    if(!empty(floatval($order_details['order']['discount_total']))){ //If we have a discount we need to add it.

        $square_order_data['order']['discounts'][0] = [
            'uid' => $order_details['order']['discount_label'],
            'name' => 'Promo Code: '.$order_details['order']['discount_label'],
            'amount_money' => [
                'currency' => $order_details['order']['order_currency'],
                'amount' => round(floatval($order_details['order']['discount_total'])*100) //needs to be in cents
            ],
            'scope' => 'ORDER'
        ];

    } //end if we have to add taxes




    if($order_details['order']['recurring_frequency']){ //it looks like this order has recurring component

    }

    // echo '<pre>';
    // print_r($square_order_data);
    // echo '</pre>';

    $square_request_body = json_encode($square_order_data);


    $curl = curl_init();

    curl_setopt_array($curl, array(
      CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$SQUARE_CONFIG['use_env']]['api_domain'].'/v2/orders',
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'POST',
      CURLOPT_POSTFIELDS => $square_request_body,
      CURLOPT_HTTPHEADER => array(
        'Square-Version: 2021-09-21',
        'Content-Type: application/json',
        'Authorization: Bearer '.$square_credentials['access_token']
      ),
    ));

    $square_create_response = curl_exec($curl);
    $info = curl_getinfo($curl);

    curl_close($curl);

    save_log('order:'.$payable_order_id, 'square_order_create', $square_request_body, $square_create_response, $info['http_code'], $info['total_time']);

    //lets convert the intent to an array.
    $order_array = json_decode($square_create_response, true);

    if(isset($order_array['order'])){
        $response['order'] = $order_array['order'];
        $response['success'] = true;
        $response['message'] = 'Created New Square Order.';
    }else if(isset($order_array['errors'])){
        $response['success'] = false;
        $response['message'] = $order_array['errors'][0]['detail'];
    }else{
        $response['success'] = false;
        $response['message'] = 'Unable to make new order with Square.';
    }


    



    if($response['success']){ //lets update this in the database

        $square_order_id = $order_array['order']['id'];
        $order_update_query = "UPDATE `orders` SET `payment_provider_order_id` = '$square_order_id', `status` = 'checkout-started', `updated` = NOW() WHERE `orders`.`id` = $internal_order_id;";
        
        if(!mysqli_query($DB_LINK, $order_update_query)){
            $response['message'] = "Failed to save Square Order in the database.";
        }

    }

    return $response;

}


function create_square_device_code(){  //This will create a Square Device Code

    global $SQUARE_CONFIG, $DB_LINK, $environment;

    $api_request_id = generate_key(20);

    $square_credentials = get_square_api_credentials();

    $square_code_data = [
      'idempotency_key' => $api_request_id,
      'device_code' => [
          'location_id' => $square_credentials['merchant']['main_location_id'],
          'product_type' => "TERMINAL_API",
          'name' => "CashEAR",
      ]
    ];



    $curl = curl_init();

    curl_setopt_array($curl, array(
      CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$SQUARE_CONFIG['use_env']]['api_domain'].'/v2/devices/codes',
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'POST',
      CURLOPT_POSTFIELDS => json_encode($square_code_data),
      CURLOPT_HTTPHEADER => array(
        'Square-Version: 2021-09-21',
        'Content-Type: application/json',
        'Authorization: Bearer '.$square_credentials['access_token']
      ),
    ));

    $square_code_response = curl_exec($curl);

    $info = curl_getinfo($curl);

    curl_close($curl);

    //lets convert the intent to an array.
    $code_array = json_decode($square_code_response, true);

    if(isset($code_array['device_code'])){
        $response['device_code'] = $code_array['device_code'];
        $response['success'] = true;
        $response['message'] = 'Created a new Device Code.';
    }else if(isset($code_array['errors'])){
        $response['success'] = false;
        $response['message'] = 'Square Said: '.$code_array['errors'][0]['detail'];
    }else{
        $response['success'] = false;
        $response['message'] = 'Unable to make new order with Square.';
    }

    return $response;

}




function get_latest_from_square($payable_order_id){  //This will get the latest on an order from square

    global $SQUARE_CONFIG, $DB_LINK, $environment;

    $response = array();
    $response['message'] = 'Initalizing';
    $response['success'] = false;

    // Step 1 is to load the order from the database
    $order_details = get_order($payable_order_id, true);
    $internal_order_id = $order_details['order']['id'];
    $square_order_id = $order_details['order']['payment_provider_order_id'];

    //Let's set a string we can use to specify the credentials.
    if($order_details['order']['test_mode']){
        $square_env_label = 'testing';
    }else{
        $square_env_label = 'production';
    }

    $response['order_environment'] = $square_env_label;

    $square_credentials = get_square_api_credentials($order_details['source']['payment_provider_connection_id'], $square_env_label);




    $curl = curl_init();

    curl_setopt_array($curl, array(
      CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$square_env_label]['api_domain'].'/v2/orders/'.$square_order_id,
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'GET',
      CURLOPT_HTTPHEADER => array(
        'Square-Version: 2021-09-21',
        'Content-Type: application/json',
        'Authorization: Bearer '.$square_credentials['access_token']
      ),
    ));

    $square_get_response = curl_exec($curl);

    $info = curl_getinfo($curl);

    curl_close($curl);

    save_log('order:'.$payable_order_id, 'square_order_get', $square_order_id, $square_get_response, $info['http_code'], $info['total_time']);

    //lets convert the intent to an array.
    $order_array = json_decode($square_get_response, true);

    if(isset($order_array['order'])){
        $response['order'] = $order_array['order'];
        $response['success'] = true;
        $response['message'] = 'Got existing Square Order.';
    }else if(isset($order_array['errors'])){
        $response['success'] = false;
        $response['message'] = 'Square Said: '.$order_array['errors'][0]['detail'];
    }else{
        $response['success'] = false;
        $response['message'] = 'Unable to make new order with Square.';
    }

    return $response;

}


function capture_order($payable_order_id, $square_payment_token){

    global $SQUARE_CONFIG, $DB_LINK, $environment;

    $response = array();
    $response['message'] = 'Initalizing';
    $response['success'] = false;

    // Step 1 is to load the order from the database
    $order_details = get_order($payable_order_id, true);
    $internal_order_id = $order_details['order']['id'];
    $square_order_id = $order_details['order']['payment_provider_order_id'];

    //Let's set a string we can use to specify the credentials.
    if($order_details['order']['test_mode']){
        $square_env_label = 'testing';
    }else{
        $square_env_label = 'production';
    }

    $response['order_environment'] = $square_env_label;

    $square_credentials = get_square_api_credentials($order_details['source']['payment_provider_connection_id'], $square_env_label);


    //So this is where we need to generate the Payment Request and associate it with the order.

    

    //TODO - Its smallest currency unit, some currencies may not need to be multiplied.
    $square_total_amount = round(floatval($order_details['order']['total'])*100); //needs to be in cents

    $api_request_id = generate_key(20);

    //lets start with the basics that will always be needed.
    $square_payment_data = [
      'idempotency_key' => $api_request_id,
      'source_id' => $square_payment_token,
      'reference_id' => $payable_order_id,
      'order_id' => $square_order_id,
      'note' => 'Payable Order ID: '.$payable_order_id,
      'accept_partial_authorization' => false, //Do not have logic in place to support this.
      'amount_money' => [
          'amount' => $square_total_amount,
          'currency' => $order_details['order']['order_currency'],
      ]
    ];


    //If we have buyer data from the checkout - we can apply to the payment object.
    if(!empty($order_details['order']['payable_buyer_data'])){

        $checkout_buyer = json_decode(base64_decode($order_details['order']['payable_buyer_data']), true);

        if(isset($checkout_buyer['buyer-email']) && !empty($checkout_buyer['buyer-email'])){
            $square_payment_data['buyer_email_address'] = $checkout_buyer['buyer-email'];
        }

        if(isset($checkout_buyer['buyer-country']) && !empty($checkout_buyer['buyer-country'])){
            $square_payment_data['billing_address']['country'] = $checkout_buyer['buyer-country'];
        }

        if(isset($checkout_buyer['buyer-name']) && !empty($checkout_buyer['buyer-name'])){
            $buyer_name_parts = explode(' ', $checkout_buyer['buyer-name'], 2);

            $square_payment_data['billing_address']['first_name'] = $buyer_name_parts[0];

            if(isset($buyer_name_parts[1])){
                $square_payment_data['billing_address']['last_name'] = $buyer_name_parts[1];
            }
        }

    }


    //If the user in a region we can add an applicaton fee, lets do that.
    if(in_array($order_details['source']['payment_provider_merchant_countrycode'], array('CA', 'US'))){

        $payable_application_fee = calculate_platform_fee($order_details['order']['total'], $order_details['order']['order_currency'], date("Y"));
        $square_application_fee = round(floatval($payable_application_fee['fee']['amount'])*100); //needs to be in cents

        $square_payment_data['app_fee_money']['amount'] = $square_application_fee;
        $square_payment_data['app_fee_money']['currency'] = $order_details['order']['order_currency'];

    }else{
        //I guess this is free.
        $payable_application_fee = 0;
    }


    $square_request_body = json_encode($square_payment_data);

    $curl = curl_init();

    curl_setopt_array($curl, array(
      CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$square_env_label]['api_domain'].'/v2/payments',
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'POST',
      CURLOPT_POSTFIELDS => $square_request_body,
      CURLOPT_HTTPHEADER => array(
        'Square-Version: 2021-09-21',
        'Content-Type: application/json',
        'Authorization: Bearer '.$square_credentials['access_token']
      ),
    ));

    $square_create_response = curl_exec($curl);
    $info = curl_getinfo($curl);

    curl_close($curl);

    save_log('order:'.$payable_order_id, 'square_payment_create', $square_request_body, $square_create_response, $info['http_code'], $info['total_time']);

    //lets convert the intent to an array.
    $payment_array = json_decode($square_create_response, true);


    if(isset($payment_array['payment']) && $payment_array['payment']['status'] == 'COMPLETED'){
        $response['success'] = true;
        $response['message'] = 'New Square Payment Created Succesfully.';
        $response['payment'] = $payment_array['payment'];

        //Pulls the latest details from Square about the Order Object.
        $squareOrder = get_latest_from_square($payable_order_id);

        //Update the Payable Order and Database.
        $update_result = mark_payable_order_as_paid_with_square($order_details, $squareOrder, $payment_array);

    }else{
        $response['success'] = false;
        if(isset($payment_array['errors'])){
            $response['message'] = $payment_array['errors'][0]['detail'];
            $response['error_code'] = $payment_array['errors'][0]['code'];
        }else{
            $response['message'] = 'Payment Failed.';
            $response['error_code'] = 'UNKNOWN';
        }
    }



    return $response;


}




//This function will finalize the order and collect all the details
function mark_payable_order_as_paid_with_square($payable_order_details, $squareOrder, $payment_array = false){

    global $SQUARE_CONFIG, $DB_LINK, $environment;

    $response = array();
    $response['message'] = 'Initalizing';
    $response['success'] = false;

    if($squareOrder['order']['state'] == 'COMPLETED'){  //lets just makr sure its in the right status.

        //most should be paid, but some non-real time payments will be in a pending status.
        $payable_order_status = 'paid';

        //So there should be a 
        $most_recent_charge = $squareOrder['order']['tenders'][0];

        //Let's Get some of the values ready for the DB Update.
        $internal_order_id = $payable_order_details['order']['id'];
        $charge_id = $most_recent_charge['id'];
        $payment_date = date("Y-m-d H:i:s", strtotime($most_recent_charge['created_at']));
        if($payment_array && isset($payment_array['payment']['app_fee_money'])){
            $platform_fee = $payment_array['payment']['app_fee_money']['amount'] / 100; //This is in cents we store it in float.
        }else{
           $platform_fee = 0; //This is in cents we store it in float. 
        }
        
        $local_fee = convert_to_local($platform_fee, $most_recent_charge['amount_money']['currency']);

        if($most_recent_charge['type'] == 'CARD'){
            $payment_method_label = 'CARD-'.strtoupper($most_recent_charge['card_details']['card']['card_brand']).'-ENDING:'.$most_recent_charge['card_details']['card']['last_4'];
        }else{
            $payment_method_label = strtoupper($most_recent_charge['type']);
        }
        
        $update_q = "UPDATE `orders` SET `status` = '$payable_order_status', `platform_fee` = '$platform_fee', `local_fee` = '$local_fee', `payment_method` = '$payment_method_label', `payment_provider_transaction_number` = '$charge_id', `payment_date` = '$payment_date', `updated` = NOW() WHERE `orders`.`id` = $internal_order_id;";

        if(mysqli_query($DB_LINK, $update_q)){
            $response['success'] = true;
            $response['message'] = "Marked the order as paid with the square data.";

            //lets also quickly update the form last payment date. 
            $source_plugin = $payable_order_details['order']['source_plugin'];
            $source_internal_id = $payable_order_details['order']['source_id'];
            $update_source_query = "UPDATE `$source_plugin` SET `last_payment` = NOW() WHERE `g_forms`.`id` = $source_internal_id;";
            mysqli_query($DB_LINK, $update_source_query);

            //For square after we have gotten this far, what if it is a subscription 
            // We will need to set that up
            if(!empty($payable_order_details['order']['recurring_frequency'])){
                //$response['subscription'] = create_recurring_order($payable_order_details);
                if($response['subscription']['success']){
                    $response['message'] .= ' Also setup a subscription.';
                }else{
                    $response['message'] .= ' But failed to setup a subscription.';
                }
            }


        }else{
            $response['message'] = "Failed to update the order in the database.";
        }


    }else{
        $response['message'] = 'Square order is not in COMPLETED State it is'+$squareOrder['order']['state'];
    }

    return $response;

}



function get_or_create_customer($customer_email){  //This will either find or create a customer


    //In this function we will lookup of the customer is here. 
    global $SQUARE_CONFIG, $DB_LINK, $environment;


    $existing_customer = check_for_customer($customer_email);



    if($existing_customer['success']){

        //Return the existing customer
        return $existing_customer;

    }else{ //we need to make a new one.



        $api_request_id = generate_key(20);

        $square_credentials = get_square_api_credentials();

        $square_customer_data = [
            'email_address' => $customer_email,
            'idempotency_key' => $api_request_id,
            'note' => "Created by CashEAR"
        ];



        $curl = curl_init();

        curl_setopt_array($curl, array(
          CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$SQUARE_CONFIG['use_env']]['api_domain'].'/v2/customers',
          CURLOPT_RETURNTRANSFER => true,
          CURLOPT_ENCODING => '',
          CURLOPT_MAXREDIRS => 10,
          CURLOPT_TIMEOUT => 0,
          CURLOPT_FOLLOWLOCATION => true,
          CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
          CURLOPT_CUSTOMREQUEST => 'POST',
          CURLOPT_POSTFIELDS => json_encode($square_customer_data),
          CURLOPT_HTTPHEADER => array(
            'Square-Version: 2021-09-21',
            'Content-Type: application/json',
            'Authorization: Bearer '.$square_credentials['access_token']
          ),
        ));

        $square_customer_response = curl_exec($curl);

        $info = curl_getinfo($curl);

        curl_close($curl);

        //lets convert the intent to an array.
        $customer_array = json_decode($square_customer_response, true);

        if(isset($customer_array['customer'])){
            $response['customer'] = $customer_array['customer'];
            $response['success'] = true;
            $response['message'] = 'Created a new customer';
        }else if(isset($customer_array['errors'])){
            $response['success'] = false;
            $response['message'] = 'Square Said: '.$customer_array['errors'][0]['detail'];
        }else{
            $response['success'] = false;
            $response['message'] = 'No customers with this email.';
        }

        return $response;


    }


}


function check_for_customer($customer_email){


    //In this function we will lookup of the customer is here. 
    global $SQUARE_CONFIG, $DB_LINK, $environment;

    $api_request_id = generate_key(20);

    $square_credentials = get_square_api_credentials();

    $square_search_data = [
      'query' => [
          'filter' => [
            'email_address' => ['exact' => $customer_email]
          ]
      ]
    ];



    $curl = curl_init();

    curl_setopt_array($curl, array(
      CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$SQUARE_CONFIG['use_env']]['api_domain'].'/v2/customers/search',
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'POST',
      CURLOPT_POSTFIELDS => json_encode($square_search_data),
      CURLOPT_HTTPHEADER => array(
        'Square-Version: 2021-09-21',
        'Content-Type: application/json',
        'Authorization: Bearer '.$square_credentials['access_token']
      ),
    ));

    $square_customer_response = curl_exec($curl);

    $info = curl_getinfo($curl);

    curl_close($curl);

    //lets convert the intent to an array.
    $customer_array = json_decode($square_customer_response, true);

    if(isset($customer_array['customers'])){
        $response['customer'] = $customer_array['customers'][0];
        $response['success'] = true;
        $response['message'] = 'Found a customer by email';
    }else if(isset($customer_array['errors'])){
        $response['success'] = false;
        $response['message'] = 'Square Said: '.$customer_array['errors'][0]['detail'];
    }else{
        $response['success'] = false;
        $response['message'] = 'No customers with this email.';
    }

    return $response;

}




function check_for_customer_group($group_name){


    //In this function we will lookup of the customer is here. 
    global $SQUARE_CONFIG, $DB_LINK, $environment;

    $api_request_id = generate_key(20);

    $square_credentials = get_square_api_credentials();


    $curl = curl_init();

    curl_setopt_array($curl, array(
      CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$SQUARE_CONFIG['use_env']]['api_domain'].'/v2/customers/groups',
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'GET',
      CURLOPT_HTTPHEADER => array(
        'Square-Version: 2021-09-21',
        'Content-Type: application/json',
        'Authorization: Bearer '.$square_credentials['access_token']
      ),
    ));

    $square_customer_response = curl_exec($curl);

    $info = curl_getinfo($curl);

    curl_close($curl);

    //lets convert the intent to an array.
    $customer_array = json_decode($square_customer_response, true);

    if(isset($customer_array['groups'])){

        //We need to loop through the groups and see if one matches.
        foreach ($customer_array['groups'] as $a_group) {
            if($a_group['name'] == $group_name){

                $response['group'] = $a_group;
                $response['success'] = true;
                $response['message'] = 'Found existing group';
                 break;
            }
        }
        if(!isset($response['group'])){
            $response['success'] = true;
            $response['message'] = 'Not found in existing groups.';
        }



    }else if(isset($customer_array['errors'])){
        $response['success'] = false;
        $response['message'] = 'Square Said: '.$customer_array['errors'][0]['detail'];
    }else{
        $response['success'] = false;
        $response['message'] = 'No groups have been setup for this merchant.';
    }

    return $response;

}



function get_or_create_customer_group($group_name){  //This will either find or create a customer group with a given name


    //In this function we will lookup of the customer is here. 
    global $SQUARE_CONFIG, $DB_LINK, $environment;


    $existing_group = check_for_customer_group($group_name);


    if($existing_group['success']){

        //Return the existing customer
        return $existing_group;

    }else{ //we need to make a new one.


        $api_request_id = generate_key(20);

        $square_credentials = get_square_api_credentials();

        $square_customer_data = [
            'group' => ['name' => $group_name],
            'idempotency_key' => $api_request_id,
        ];



        $curl = curl_init();

        curl_setopt_array($curl, array(
          CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$SQUARE_CONFIG['use_env']]['api_domain'].'/v2/customers/groups',
          CURLOPT_RETURNTRANSFER => true,
          CURLOPT_ENCODING => '',
          CURLOPT_MAXREDIRS => 10,
          CURLOPT_TIMEOUT => 0,
          CURLOPT_FOLLOWLOCATION => true,
          CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
          CURLOPT_CUSTOMREQUEST => 'POST',
          CURLOPT_POSTFIELDS => json_encode($square_customer_data),
          CURLOPT_HTTPHEADER => array(
            'Square-Version: 2021-09-21',
            'Content-Type: application/json',
            'Authorization: Bearer '.$square_credentials['access_token']
          ),
        ));

        $square_customer_response = curl_exec($curl);

        $info = curl_getinfo($curl);

        curl_close($curl);

        //lets convert the intent to an array.
        $customer_array = json_decode($square_customer_response, true);

        if(isset($customer_array['group'])){
            $response['group'] = $customer_array['group'];
            $response['success'] = true;
            $response['message'] = 'Created a new customer group.';
        }else if(isset($customer_array['errors'])){
            $response['success'] = false;
            $response['message'] = 'Square Said: '.$customer_array['errors'][0]['detail'];
        }else{
            $response['success'] = false;
            $response['message'] = 'No customers with this email.';
        }

        return $response;


    }


}


//This function will handle the complexity involved in getting/creating a customer and
// assign that customer to a given group.
function put_customer_in_group($square_customer_id, $group_name){

    $square_group = get_or_create_customer_group($group_name);

    //In this function we will lookup of the customer is here. 
    global $SQUARE_CONFIG, $DB_LINK, $environment;

    $square_credentials = get_square_api_credentials();

    $curl = curl_init();

    curl_setopt_array($curl, array(
      CURLOPT_URL => 'https://'.$SQUARE_CONFIG[$SQUARE_CONFIG['use_env']]['api_domain'].'/v2/customers/'.$square_customer_id.'/groups/'.$square_group['group']['id'],
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'PUT',
      CURLOPT_HTTPHEADER => array(
        'Square-Version: 2021-09-21',
        'Content-Type: application/json',
        'Authorization: Bearer '.$square_credentials['access_token']
      ),
    ));

    $square_customer_response = curl_exec($curl);

    $info = curl_getinfo($curl);

    curl_close($curl);

    //lets convert the intent to an array.
    $customer_array = json_decode($square_customer_response, true);

    if(isset($customer_array['customers'])){
        $response['customer'] = $customer_array['customers'][0];
        $response['success'] = true;
        $response['message'] = 'Found a customer by email';
    }else if(isset($customer_array['errors'])){
        $response['success'] = false;
        $response['message'] = 'Square Said: '.$customer_array['errors'][0]['detail'];
    }else{
        $response['success'] = false;
        $response['message'] = 'No customers with this email.';
    }

    return $response;




}





function check_for_payments_in_last($seconds, $location_id = false){


    //In this function we will lookup of the customer is here. 
    global $SQUARE_CONFIG, $DB_LINK, $environment;

    $api_request_id = generate_key(20);

    $square_credentials = get_square_api_credentials();

    //If it was not passed in lets use the default one.
    if(!$location_id){
        $location_id = $square_credentials['merchant']['main_location_id'];
    }


    date_default_timezone_set('UTC');
    $begin_time = date('Y-m-d\TH:i:s', strtotime('-'.$seconds.' seconds'));
    $end_time = date('Y-m-d\TH:i:s', strtotime('now'));

    $response['begin_time'] = $begin_time;
    $response['end_time'] = $end_time;


    $url_get = 'https://'.$SQUARE_CONFIG[$SQUARE_CONFIG['use_env']]['api_domain'].'/v2/payments?location_id='.$location_id.'&begin_time='.$begin_time.'&end_time='.$end_time;

    $curl = curl_init();

    curl_setopt_array($curl, array(
      CURLOPT_URL => $url_get,
      CURLOPT_RETURNTRANSFER => true,
      CURLOPT_ENCODING => '',
      CURLOPT_MAXREDIRS => 10,
      CURLOPT_TIMEOUT => 0,
      CURLOPT_FOLLOWLOCATION => true,
      CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
      CURLOPT_CUSTOMREQUEST => 'GET',
      CURLOPT_HTTPHEADER => array(
        'Square-Version: 2021-09-21',
        'Content-Type: application/json',
        'Authorization: Bearer '.$square_credentials['access_token']
      ),
    ));

    $square_payments_response = curl_exec($curl);

    $info = curl_getinfo($curl);

    curl_close($curl);

    //lets convert the intent to an array.
    $payments_array = json_decode($square_payments_response, true);

    if(isset($payments_array['payments'])){

        $response['payments'] = $payments_array['payments'];
        $response['success'] = true;
        $response['message'] = 'Found payments for the time period';

    }else if(isset($customer_array['errors'])){
        $response['success'] = false;
        $response['message'] = 'Square Said: '.$customer_array['errors'][0]['detail'];
    }else{
        $response['success'] = false;
        $response['message'] = 'No payments found for the time period.';
    }

    return $response;

}




//This function will check and report on tips in the las X seconds
function check_for_tips_in_last($seconds){

    //This is a baseline number for generosity detection
    //In the future you could make it detectible.
    $tip_generosity_baseline = 0.15;

    //These are some sample BINs that signify someone is high income
    //https://www.freebinchecker.com/credit-card-bin-search/ultra-high-net-worth-brand
    $ultra_high_income_bins = array('445239', '445239', '445239');

    //These are some basic student bins
    //https://www.freebinchecker.com/canada_unembossed-prepaid-student-bin-list-brand
    $basic_income_bins = array('539847', '545722', '539847');

    $response['success'] = false;
    $response['tips'] = array();
    
    $square_payments = check_for_payments_in_last($seconds);

   if($square_payments['success']){


    //Lets loop through each of the payments and investigat the tip level
    foreach($square_payments['payments'] as $a_payment){

        //Thats a good sign there might be a tip. 
        if($a_payment['status'] == 'COMPLETED' && isset($a_payment['tip_money'])){

            $response['success'] = true;

            $total_amount = $a_payment['amount_money']['amount'];
            $tip_amount = $a_payment['tip_money']['amount'];
            $tip_percentage = round($tip_amount/$total_amount, 2);

            //Now we need to look at the BIN Data if available to help sqew generocity.
            if(isset($a_payment['card_details']) && isset($a_payment['card_details']['card'])){
                $card_bin = $a_payment['card_details']['card']['bin'];

                //This is where we could use logic to tweak the baseline based on card BINs
                if(in_array($card_bin, $ultra_high_income_bins)){
                    $generocity_baseline = $tip_generosity_baseline + 0.05;
                }else if(in_array($card_bin, $basic_income_bins)){
                    $generocity_baseline = $tip_generosity_baseline - 0.05;
                }else{
                    $generocity_baseline = $tip_generosity_baseline;  
                }
            }else{
                $generocity_baseline = $tip_generosity_baseline;
            }

            if($tip_percentage >  $generocity_baseline){
                $tip_type = 'generous';

                //We need to save our generous customers to our generous customer groups.
                if(isset($a_payment['customer_id']) && !empty($a_payment['customer_id'])){

                    put_customer_in_group($a_payment['customer_id'], 'generous');

                }

            }else{
                $tip_type = 'normal';
            }


            array_push($response['tips'], array('type'=>$tip_type, 'amount'=> $tip_amount, 'percentage' => $tip_percentage, 'baseline' => $generocity_baseline));

        }

        return $response;

    }

   }else{

    return $response;

   }


}


